#include "main.h"

// Tiny sample program
void sample(){
	// Total 68 bytes
	asm volatile("addiu       $s0,$zero,0x49");
	asm volatile("lui         $s1,0xA000");
	asm volatile("addiu       $s2,$zero,0x00");
	asm volatile("addiu       $s3,$zero,0x00");
	asm volatile("addu        $s2,$s2,$s1");
	asm volatile("sb          $s0,0x1FF8($s2)");
	asm volatile("nop; jal    0x90000750");
	asm volatile("addiu       $s3,$s3,1");
	asm volatile("andi        $s3,$s3,15");
	asm volatile("nop;bne     $s3,$zero,-20");
	asm volatile("sb          $zero,0x1FF8($s2)");
	asm volatile("addiu       $s2,$s2,1");
	asm volatile("andi        $s2,$s2,7");
	asm volatile("nop; b      -48");
}

// 0xA0000700 RGDSP Display registers
// 0xA0000710 KEYIN Key input
// 0xA0000720 INPUT Key input (no waint)
// 0xA0000730 D1 4.5 msec timer
// 0xA0000740 D2 9 msec timer
// 0xA0000750 D3 27 msec timer
const unsigned char segdata[]={
	0x5C,0x06,0x5B,0x4F,
	0x66,0x6D,0x7D,0x27,
	0x7F,0x6F,0x77,0x7C,
	0x39,0x5E,0x79,0x71
};
void subRGDSP(unsigned long reg){
	unsigned char b;
	unsigned long i;
	for (i=0;i<8;i++) {
		RAM[0x17FF-i]=segdata[reg&0xF];
		reg=reg>>4;
	}
}
void subD1(){
	unsigned long start,end;
	start=g_TMR1h|TMR1;
	if (start<0xFFFFFFFFFFFD400) {
		end=start+0x2c000;
		while((g_TMR1h|TMR1)<end) {
			if ((g_TMR1h|TMR1)<start) break;
		}
	} else {
		end=start-0xFFFFFFFFFFFD400;
		while(start<=(g_TMR1h|TMR1));
		while((g_TMR1h|TMR1)<end);
	}
}
void subD2(){
	subD1();
	subD1();
}
void subD3(){
	subD2();
	subD2();
	subD2();
}
long subINPUT(){
	unsigned long i;
	for (i=0;i<8;i++) {
		if (!(g_keymatrix[0]&(1<<i))) return i;
	}
	for (i=0;i<8;i++) {
		if (!(g_keymatrix[1]&(1<<i))) return 0x08+i;
	}
	for (i=0;i<8;i++) {
		if (!(g_keymatrix[2]&(1<<i))) return 0x10+i;
	}
	return 0xff;
}
long subKEYIN(){
	unsigned long ret;
	subD3(); // Anti chattering
	while (subINPUT()<0xff);
	subD3(); // Anti chattering
	do {
		ret=subINPUT();
	} while (ret==0xff);
	return ret; 
}

// Template routine used in addresses 0xA0000700 etc.
void jumpasm(){
	asm volatile("lui   $v0,0x1234");
	asm volatile("addiu $v0,$v0,0x5678");
	asm volatile("jr    $v0");
}
// Modify the rutine used in addresses 0xA0000700 etc.
void init_mon_sub(unsigned long pos, void* addr){
	SUBS[pos++]=0xff&(((unsigned long)addr)>>16);
	SUBS[pos++]=0xff&(((unsigned long)addr)>>24);
	pos++;
	pos++;
	SUBS[pos++]=0xff&((unsigned long)addr);
	SUBS[pos++]=0xff&(((unsigned long)addr)>>8);
}
void init_mon(){
	unsigned long i;
	// Construct jump routines in RAM area.
	for (i=0;i<12;i++) {
		SUBS[i+0x00]=((char*)jumpasm)[i];
		SUBS[i+0x10]=((char*)jumpasm)[i];
		SUBS[i+0x20]=((char*)jumpasm)[i];
		SUBS[i+0x30]=((char*)jumpasm)[i];
		SUBS[i+0x40]=((char*)jumpasm)[i];
		SUBS[i+0x50]=((char*)jumpasm)[i];
	}
	// 0xA0000700 subRGDSP
	init_mon_sub(0x00,subRGDSP);
	// 0xA0000710 subKEYIN
	init_mon_sub(0x10,subKEYIN);
	// 0xA0000720 subINPUT
	init_mon_sub(0x20,subINPUT);
	// 0xA0000730 subD1
	init_mon_sub(0x30,subD1);
	// 0xA0000740 subD2
	init_mon_sub(0x40,subD2);
	// 0xA0000750 subD3
	init_mon_sub(0x50,subD3);

	// Load test program just after power-on.
	if (RCONbits.POR){
		RCONbits.POR=0;
		for(i=0;i<(sizeof RAM);i++){
			if (i<68) {
				RAM[i]=((const unsigned char*)sample)[i];
			} else {
				RAM[i]=0;
			}
		}
	}
}

unsigned long read(unsigned long reg) {
	unsigned long i;
	reg=reg&0x1FFF0000;
	i=reg>>16;
	i=i&0x1FFF;
	if (0x801<=i) reg|=((unsigned long)RAM[i-0x801])<<8;
	if (0x800<=i) reg|=RAM[i-0x800];
	return reg;
}
void write(unsigned long reg) {
	unsigned long i;
	reg=reg&0x1FFF00FF;
	i=reg>>16;
	i=i&0x1FFF;
	if (0x800<=i) RAM[i-0x800]=reg&0xff;
}
void run(unsigned long addr){
	addr=addr>>16;
	if (addr<0x800 || 0x1ffc<addr) return;
	addr&=0x1ffc;
	void (*f)()=0xA0000000|addr;
	asm volatile("addiu $sp,$sp,-40");
	asm volatile("sw    $s8,36($sp)");
	asm volatile("sw    $s7,32($sp)");
	asm volatile("sw    $s6,28($sp)");
	asm volatile("sw    $s5,24($sp)");
	asm volatile("sw    $s4,20($sp)");
	asm volatile("sw    $s3,16($sp)");
	asm volatile("sw    $s2,12($sp)");
	asm volatile("sw    $s1,8($sp)");
	asm volatile("sw    $s0,4($sp)");
	f();
	asm volatile("lw    $s0,4($sp)");
	asm volatile("lw    $s1,8($sp)");
	asm volatile("lw    $s2,12($sp)");
	asm volatile("lw    $s3,16($sp)");
	asm volatile("lw    $s4,20($sp)");
	asm volatile("lw    $s5,24($sp)");
	asm volatile("lw    $s6,28($sp)");
	asm volatile("lw    $s7,32($sp)");
	asm volatile("lw    $s8,36($sp)");
	asm volatile("addiu $sp,$sp,40");
}

// Main monitor loop
void monitor() {
	unsigned long reg,i;
	reg=0;
	while(1){
		// Show in 7-segment LED display
		subRGDSP(reg);
		// Key input and responces.
		i=subKEYIN();
		switch (i) {
			case 0x10: // RUN
				run(reg);
				break;
			case 0x11: // RET
				break;
			case 0x12: // ADRS SET
				reg=(reg&0x1fff)<<16;
				reg=read(reg);
				break;
			case 0x13: // READ DECR
				reg-=0x10000;
				reg=read(reg);
				break;
			case 0x14: // READ INCR
				reg+=0x10000;
				reg=read(reg);
				break;
			case 0x15: // WRITE INCR
				write(reg);
				reg+=0x10000;
				reg=read(reg);
				break;
			case 0x16: // STORE DATA
				break;
			case 0x17: // LOAD DATA
				break;
			default:   // Numeric keys
				reg=(reg&0xFFFF0000)|((reg&0xFFF)<<4)|(i&0xF);
				break;
		}
	}
}
